from cStringIO import StringIO
from nose.tools import assert_equal, assert_less, assert_less_equal, assert_not_equal

from casava import reader_impl


def test_easy():
    fp = StringIO('really,easy,data\nis,really,easy')
    result = list(reader_impl.reader(fp))
    assert_equal(result, [[u'really',u'easy',u'data'], [u'is',u'really',u'easy']])


def test_semi():
    fp = StringIO('really;easy;data\nis;really;easy')
    result = list(reader_impl.reader(fp))
    assert_equal(result, [[u'really',u'easy',u'data'], [u'is',u'really',u'easy']])


def test_tab():
    fp = StringIO('really\teasy\tdata\nis\treally\teasy')
    result = list(reader_impl.reader(fp))
    assert_equal(result, [[u'really',u'easy',u'data'], [u'is',u'really',u'easy']])


def test_cr():
    fp = StringIO('really,easy,data\ris,really,easy')
    result = list(reader_impl.reader(fp))
    assert_equal(result, [[u'really',u'easy',u'data'], [u'is',u'really',u'easy']])


def test_crlf():
    fp = StringIO('really,easy,data\r\nis,really,easy')
    result = list(reader_impl.reader(fp))
    assert_equal(result, [[u'really',u'easy',u'data'], [u'is',u'really',u'easy']])


def test_quoted_newlines():
    fp = StringIO('really,"easy\n",data\r\nis,really,easy')
    result = list(reader_impl.reader(fp))
    assert_equal(result, [[u'really',u'easy\n',u'data'], [u'is',u'really',u'easy']])


def test_weird_unicode():
    zh_data = u'\u5e03\u4f9d\u65cf\u82d7\u65cf'
    lines = [
        '"{}","{}"'.format(zh_data.encode('ascii', 'ignore'), ""),
        '"{}","{}"'.format(zh_data.encode('utf-8'), zh_data.encode('utf-16')),
        ]
    fp = StringIO('\r\n'.join(lines))
    result = list(reader_impl.reader(fp))
    print result

# Test byte accumulation
def test_acc_bytes_small():
    lines = 'This is a set of lines of varying length'.split()
    lines = [line + '\n' for line in lines]
    it = iter(lines)
    prefix = reader_impl.accumulate_bytes(it, 20)
    assert_less_equal(20, len(prefix))
    assert_equal(prefix + ''.join(list(it)), ''.join(lines))


def test_acc_bytes_large():
    lines = 'This is a set of lines of varying length'.split()
    lines = [line + '\n' for line in lines]
    it = iter(lines)
    prefix = reader_impl.accumulate_bytes(it, 200)
    assert_less_equal(len(prefix), 200)
    assert_equal([], list(it))
    assert_equal(prefix, ''.join(lines))


# Test auto unicode
def test_auto_unicode():
    zh_utf16_bytes = '\xff\xfe\x03^\x9dO\xcfe\xd7\x82\xcfe'
    zh_uni = zh_utf16_bytes.decode('utf-16LE')
    assert_equal(reader_impl.auto_unicode(zh_utf16_bytes), zh_uni)
    zh_utf16_bytes_bad = '\xff\xfex03^\x9dOxcfe\xd7\x82\xcfe'
    assert_not_equal(reader_impl.auto_unicode(zh_utf16_bytes_bad), zh_uni)





def _chunk(iterator, chunksize):
    chunk = []
    iterator = iter(iterator)
    while True:
        try:
            chunk.append(iterator.next())
        except StopIteration:
            if chunk:
                yield chunk
            break
        if len(chunk) >= chunksize:
            yield chunk
            chunk = []





import csv
import logging
from itertools import chain

import chardet


log = logging.getLogger(__name__)


class reader(object):
    ENC_DETECTION_SIZE = 100 * 1024
    SEP_DETECTION_SIZE = 1024
    SEP_CHARS = [',', ';', '\t']

    def __init__(
            self,
            content_iter,
            enc_detection_size=None,
            sep_detection_size=None):
        self.content_iter = content_iter
        self._enc_detection_size = enc_detection_size or self.ENC_DETECTION_SIZE
        self._sep_detection_size = sep_detection_size or self.SEP_DETECTION_SIZE

    def __iter__(self):
        encoding, dialect = self._detect_encoding_dialect()
        log.info('detect encoding: %r', encoding)
        log.info('detect eol     : %r', dialect.lineterminator)
        log.info('detect delim   : %r', dialect.delimiter)
        safe_content_iter = safe_for_ilines(self.content_iter)
        cur_line_iter = ilines(safe_content_iter)
        rdr = csv.reader(cur_line_iter, dialect)
        lineno = 0
        try:
            while True:
                lineno += 1
                try:
                    row = rdr.next()
                    yield self._decode_row(row, encoding)
                except StopIteration:
                    raise
                except Exception:
                    log.exception('Error on line %d', lineno)
        except StopIteration:
            pass

    def _detect_encoding_dialect(self):
        content_header = accumulate_bytes(self.content_iter, self._enc_detection_size)
        if not isinstance(content_header, unicode):
            encoding = chardet.detect(content_header)
        else:  # CSV module expects bytes, so we force it.
            encoding = dict(encoding='utf8')
            content_header = content_header.encode('utf8')
            self.content_iter = (s.encode('utf8') for s in self.content_iter)
        try:
            dialect = csv.Sniffer().sniff(content_header, delimiters=',;|\t\x1f')
        except csv.Error:
            dialect = csv.excel
        self.content_iter = chain([content_header], self.content_iter)
        return encoding['encoding'], dialect

    def _decode_row(self, row, encoding):
        result = []
        if encoding is None:
            return [auto_unicode(b_cell) for b_cell in row]
        for b_cell in row:
            try:
                u_cell = unicode(b_cell, encoding)
            except UnicodeDecodeError:
                u_cell = auto_unicode(b_cell)
            result.append(u_cell)
        return result


def auto_unicode(bytes):
    '''Guaranteed to return valid unicode string'''
    enc = chardet.detect(bytes)
    if enc['encoding']:
        return bytes.decode(enc['encoding'], 'ignore')
    else:
        return bytes.decode('utf-8', 'ignore')


def safe_for_ilines(source_iterable, block_size=8192):
    '''Yield blocks of an approximate size that *never* end with \r or \n'''
    tail = ''
    source_gen = iter(source_iterable)
    while True:
        while len(tail) > block_size:
            endpos = block_size
            while endpos and tail[endpos-1] in ('\r', '\n'):
                endpos -= 1
            if endpos == 0:
                # Nothing but \r\n
                break
            to_yield, tail = tail[:endpos], tail[endpos:]
            yield to_yield
        try:
            tail += source_gen.next()
        except StopIteration:
            break
    yield tail


# From http://code.activestate.com/recipes/286165-ilines-universal-newlines-from-any-data-source/
def ilines(source_iterable):
    '''yield lines as in universal-newlines from a stream of data blocks'''
    tail = ''
    for block in source_iterable:
        if not block:
            continue
        if tail.endswith('\r'):
            yield tail[:-1] + '\n'
            if block.startswith('\n'):
                pos = 1
            else:
                tail = ''
        else:
            pos = 0
        try:
            while True: # While we are finding LF.
                npos = block.index('\n', pos) + 1
                try:
                    rend = npos - 2
                    rpos = block.index('\r', pos, rend)
                    if pos:
                        yield block[pos : rpos] + '\n'
                    else:
                        yield tail + block[:rpos] + '\n'
                    pos = rpos + 1
                    while True: # While CRs 'inside' the LF
                        rpos = block.index('\r', pos, rend)
                        yield block[pos : rpos] + '\n'
                        pos = rpos + 1
                except ValueError:
                    pass
                if '\r' == block[rend]:
                    if pos:
                        yield block[pos : rend] + '\n'
                    else:
                        yield tail + block[:rend] + '\n'
                elif pos:
                    yield block[pos : npos]
                else:
                    yield tail + block[:npos]
                pos = npos
        except ValueError:
            pass
        # No LFs left in block.  Do all but final CR (in case LF)
        try:
            while True:
                rpos = block.index('\r', pos, -1)
                if pos:
                    yield block[pos : rpos] + '\n'
                else:
                    yield tail + block[:rpos] + '\n'
                pos = rpos + 1
        except ValueError:
            pass

        if pos:
            tail = block[pos:]
        else:
            tail += block
    if tail:
        yield tail


def accumulate_bytes(it, size):
    '''Accumulate around `size` bytes from the content iterator'''
    cur_buf = []
    cur_size = 0
    try:
        while cur_size < size:
            chunk = it.next()
            cur_buf.append(chunk)
            cur_size += len(chunk)
    except StopIteration:
        pass
    return ''.join(cur_buf)



# Copyright 2013-2014 Synappio LLC. All rights reserved.
import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()

setup(
    name='casava',
    version='0.0',
    description='casava: a csv reader thingy',
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
    ],
    author='Rick Copeland',
    author_email='rick@arborian.com',
    maintainer='Rick Copeland',
    maintainer_email='rick@arborian.com',
    url='https://github.com/synappio/casava',
    keywords='csv',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'chardet>=2.2.1',
    ],
    tests_require=[],
    test_suite="casava",
    entry_points="""\
    """,
)
